from astoptimizer.compatibility import FLOAT_TYPES, STR_TYPES, COMPLEX_TYPES

DEFAULT_FEATURES = ('builtins', 'math')

def check_args(min_narg, max_narg, *arg_types):
    """
    Callback checking function arguments for Config.pure_functions.
    ensure that the function has betwen min_narg and max_narg arguments, and
    arguments types conform to the specified types.
    """
    def check(config, args):
        if not(min_narg <= len(args) <= max_narg):
            return False
        for index, arg in enumerate(args):
            types = arg_types[index]
            if not isinstance(arg, types):
                return False
        return True
    return check

def check_1arg(types):
    """
    Callback checking function arguments for Config.pure_functions:
    ensure that the function has exactly one argument and the argument type
    is one type of types.
    """
    return check_args(1, 1, types)

def check_noarg(config, args):
    return len(args) == 0

def check_method_noarg(config, obj, args):
    return len(args) == 0

class Config:
    def __init__(self, *features):
        self.features = set(DEFAULT_FEATURES + features)

        # Hardcoded constants
        self.constants = {
            'True': True,
            'False': False,
        }

        # List of functions that can be called at compile time (no border
        # effect), dictionnary of name => (check_callback, func) with
        # check_callback returning False if the call cannot be done:
        #
        # def check_callback(config, args):
        #    if ...:
        #        # can be done at compile time
        #        return True
        #    else:
        #        # must be done at runtime
        #        return False
        self.pure_functions = {}

        # List of methods that can be called at compile time (no border
        # effect), dictionnary of type => {attr_name => (check_callback, func)}
        # with check_callback returning False if the call cannot be done:
        #
        # def check_callback(config, obj, args):
        #    if ...:
        #        # can be done at compile time
        #        return True
        #    else:
        #        # must be done at runtime
        #        return False
        self.pure_methods = {}

        # Don't optimize on operation on bytes/str if the result is longer than
        # max_string_length characters/bytes
        self.max_string_length = 4096

        # Don't optimize on operation on integers if the results if not in
        # range [min_number; max_number]. For complex numbers, check if
        # abs(result) is in the range [min_number; max_number].
        self.min_number = -2**100
        self.max_number = 2**100

        for feature in self.features:
            modname = "config_%s" % feature
            mod = __import__("astoptimizer.%s" % modname)
            mod = getattr(mod, modname)
            mod.setup_config(self)

    def valid_result(self, result):
        if isinstance(result, FLOAT_TYPES):
            if not (self.min_number <= result <= self.max_number):
                return False
        elif isinstance(result, STR_TYPES):
            if len(result) > self.max_string_length:
                return False
        elif isinstance(result, COMPLEX_TYPES):
            if not (self.min_number <= abs(result) <= self.max_number):
                return False
        return True

