"""
Enable optimizations on the builtins functions (ex: len()) and types (ex:
unicode.encode()).
"""

from astoptimizer.compatibility import (
    PYTHON3,
    check_unicode_subscript,
    INT_TYPES, FLOAT_TYPES, COMPLEX_TYPES,
    NATIVE_STR, BYTES_TYPE, UNICODE_TYPE, STR_TYPES)
import sys
from astoptimizer.config import check_args, check_method_noarg, check_1arg

def check_ord_args(config, args):
    if len(args) != 1:
        return False
    arg = args[0]
    if not isinstance(arg, STR_TYPES):
        return False
    return len(arg) == 1

def check_strfunc_args(func):
    def check_args(config, args):
        if len(args) != 1:
            return False
        arg = args[0]
        if not isinstance(arg, STR_TYPES + COMPLEX_TYPES):
            return False
        result = func(arg)
        if len(result) > config.max_string_length:
            return False
        return True
    return check_args

if not PYTHON3:
    def check_unicode_args(config, args):
        if len(args) != 1:
            return False
        arg = args[0]
        if not isinstance(arg, STR_TYPES + COMPLEX_TYPES):
            return False
        if isinstance(arg, BYTES_TYPE):
            if any(ord(char) > 127 for char in arg):
                return False
        result = unicode(arg)
        if len(result) > config.max_string_length:
            return False
        return True

def check_encoding(encoding):
    if not isinstance(encoding, NATIVE_STR):
        return False
    # FIXME: allow more codecs?
    if encoding not in ('ascii', 'latin1', 'utf8', 'utf-8'):
        return False
    return True

def check_decode_args(config, obj, args):
    if len(args) not in (1, 2):
        return False
    if not check_encoding(args[0]):
        return False
    if len(args) == 2:
        errors = args[1]
        if not isinstance(errors, NATIVE_STR):
            return False
        # FIXME: allow more error handlers?
        if errors not in ('strict', 'replace', 'ignore'):
            return False
    return True

def check_encode_args(config, obj, args):
    if len(args) not in (1, 2):
        return False
    if not check_encoding(args[0]):
        return False
    if len(args) == 2:
        errors = args[1]
        if not isinstance(errors, NATIVE_STR):
            return False
        # FIXME: allow more error handlers?
        if errors not in ('strict', 'replace', 'ignore'):
            return False
    if not check_unicode_subscript(config, obj):
        return False
    return True

def setup_config(config):
    config.pure_functions.update({
        'abs': (check_1arg(COMPLEX_TYPES), abs),
        'bool': (check_1arg(FLOAT_TYPES + STR_TYPES), bool),
        'chr': (check_1arg(INT_TYPES), chr),
        'float': (check_1arg(INT_TYPES + (float,)), float),
        'int': (check_1arg(INT_TYPES + (float,)), int),
        'len': (check_1arg(STR_TYPES), len),
        'oct': (check_1arg(INT_TYPES), oct),
        'ord': (check_ord_args, ord),
        'pow': (check_args(2, 3, FLOAT_TYPES, FLOAT_TYPES, FLOAT_TYPES), pow),
        'repr': (check_strfunc_args(repr), repr),
        'round': (check_args(1, 2, FLOAT_TYPES, INT_TYPES), round),
        'str': (check_strfunc_args(str), str),
    })
    if sys.version_info >= (2,6):
        config.pure_functions['bin'] = (check_1arg(INT_TYPES), bin)
    if not PYTHON3:
        config.pure_functions.update({
            'long': (check_1arg(INT_TYPES + (float,)), int),
            'unichr': (check_1arg(INT_TYPES), unichr),
            'unicode': (check_unicode_args, unicode),
        })

    config.pure_methods.update({
        BYTES_TYPE: {'decode': check_decode_args},
        UNICODE_TYPE: {'encode': check_encode_args},
        int: {'bit_length': check_method_noarg},
    })
    if not PYTHON3:
        config.pure_methods[long] = config.pure_methods[int]

