import operator
from astoptimizer import UNSET
from astoptimizer.config import Config
from astoptimizer.compatibility import (
    ast,
    PYTHON3,
    check_unicode_subscript,
    INT_TYPES, FLOAT_TYPES, COMPLEX_TYPES,
    NATIVE_STR, BYTES_TYPE, UNICODE_TYPE, STR_TYPES)
import sys

EVAL_COMPARE = {
    # a == b, a != b
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    # a < b, a <= b
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    # a > b, a >= b
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    # a in b
    ast.In: operator.contains,
}

EVAL_UNARYOP = {
    # not a, ~a, -a
    ast.Not: operator.not_,
    ast.Invert: operator.invert,
    ast.USub: operator.neg,
}

EVAL_BINOP = {
    # a + b, a - b, a * b
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.FloorDiv: operator.floordiv,
    # a ** b
    ast.Pow: operator.pow,
    # a << b, a >> b
    ast.LShift: operator.lshift,
    ast.RShift: operator.rshift,
    # a & b, a | b, a ^ b
    ast.BitAnd: operator.and_,
    ast.BitOr: operator.or_,
    ast.BitXor: operator.xor,
}
if PYTHON3:
    EVAL_BINOP[ast.Div] = operator.truediv
else:
    if sys.version_info >= (2,6) and sys.flags.division_new:
        # -Qnew
        EVAL_BINOP[ast.Div] = operator.truediv
    else:
        # -Qold or Python < 2.7
        EVAL_BINOP[ast.Div] = operator.floordiv

EVAL_BOOLOP = {
    # a and b, a or b
    ast.And: operator.__and__,
    ast.Or: operator.__or__,
}

class Optimizer(ast.NodeTransformer):
    def __init__(self, config=None):
        ast.NodeTransformer.__init__(self)
        if config is None:
            config = Config()
        self.config = config
        self.variables = {}

    def fix_node(self, node, new_node):
        new_node.lineno = 1
        new_node.col_offset = 1
        ast.fix_missing_locations(new_node)
        return ast.copy_location(new_node, node)

    def get_constant(self, node):
        if node is None:
            return None
        if isinstance(node, ast.Num):
            return node.n
        if PYTHON3:
            if isinstance(node, ast.Str):
                return node.s
            if isinstance(node, ast.Bytes):
                return node.s
        else:
            if isinstance(node, ast.Str):
                return node.s
        if isinstance(node, ast.Name):
            return self.config.constants.get(node.id, UNSET)
        return UNSET

    def get_constant_list(self, nodes):
        constants = []
        for node in nodes:
            constant = self.get_constant(node)
            if constant is UNSET:
                return None
            constants.append(constant)
        return constants

    def new_constant(self, node, value):
        if isinstance(value, bool):
            name = "True" if value else "False"
            new_node = ast.Name(id=name, ctx=ast.Load())
        elif isinstance(value, COMPLEX_TYPES):
            new_node = ast.Num(n=value)
        elif isinstance(value, UNICODE_TYPE):
            if PYTHON3:
                new_node = ast.Str(s=value)
            else:
                new_node = ast.Str(s=value)
        elif isinstance(value, BYTES_TYPE):
            if PYTHON3:
                new_node = ast.Bytes(s=value)
            else:
                new_node = ast.Str(s=value)
        elif value is None:
            new_node = ast.Name(id="None", ctx=ast.Load())
        else:
            raise NotImplementedError("unable to create an AST object for constant: %r" % (value,))
        return self.fix_node(node, new_node)

    def do_visit_list(self, nodes):
        return [self.visit(expr) for expr in nodes]

    def visit(self, node):
        new_node = ast.NodeTransformer.visit(self, node)
        if new_node is None:
            return node
        return new_node

    def visit_Name(self, node):
        name = node.id
        constant = self.config.constants.get(name, UNSET)
        if constant is UNSET:
            return
        return self.new_constant(node, constant)

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name):
            fullname = "%s.%s" % (node.value.id, node.attr)
            constant = self.config.constants.get(fullname, UNSET)
            if constant is UNSET:
                return
            return self.new_constant(node, constant)
        return

    def binop(self, node, eval_binop, left, left_cst, right, right_cst):
        # FIXME: support complex, issue with ?? operator
        if left_cst is not UNSET and right_cst is not UNSET:
            ok = (isinstance(left_cst, INT_TYPES)
                  and isinstance(right_cst, INT_TYPES))
            if (not ok
            and isinstance(left_cst, COMPLEX_TYPES)
            and isinstance(right_cst, COMPLEX_TYPES)):
                ok = isinstance(node.op,
                        (ast.Add, ast.Sub,
                         ast.Mult, ast.Div, ast.FloorDiv,
                         ast.Pow))
            if not ok and isinstance(node.op, ast.Mult):
                ok = ((isinstance(left_cst, STR_TYPES)
                       and isinstance(right_cst, INT_TYPES))
                      or (isinstance(left_cst, INT_TYPES)
                          and isinstance(right_cst, STR_TYPES)))
            if not ok:
                return UNSET
            result = eval_binop(left_cst, right_cst)
            if not self.config.valid_result(result):
                return UNSET
            return self.new_constant(node, result)

        if left_cst is not UNSET:
            # x + 0 -> x
            # FIXME: False + 0 -> 0, not False
            if (isinstance(node.op, ast.Add)
            and left_cst == 0
            and isinstance(left_cst, int)):
                return right
            # x * 1 -> x
            # FIXME: True * 1 -> 1, not True
            if (isinstance(node.op, ast.Mult)
            and left_cst == 1
            and isinstance(left_cst, int)):
                return right

        if right_cst is not UNSET:
            # O + x -> x
            # FIXME: 0 + False -> 0, not False
            if (isinstance(node.op, ast.Add)
            and right_cst == 0
            and isinstance(right_cst, int)):
                return left
            # 1 * x -> x
            # FIXME: 1 * True -> 1, not True
            if (isinstance(node.op, ast.Mult)
            and right_cst == 1
            and isinstance(right_cst, int)):
                return left

        return UNSET

    def visit_BinOp(self, node):
        eval_binop = EVAL_BINOP.get(node.op.__class__)
        if not eval_binop:
            return
        if not PYTHON3 and isinstance(node.op, ast.Div):
            # FIXME: allow also this optimization if
            # "from __future_ import division"
            # is used: because it might be affected at runtime by the -Q
            # flag.
            # self.compile_info.flags & consts.CO_FUTURE_DIVISION:
            # a / b, a // b
            if not 'pythonenv' in self.config.features:
                return
        left = self.visit(node.left)
        left_cst = self.get_constant(left)
        right = self.visit(node.right)
        right_cst = self.get_constant(right)
        new_node = self.binop(node, eval_binop, left, left_cst, right, right_cst)
        if new_node is not UNSET:
            return new_node
        return ast.BinOp(left, node.op, right)

    def visit_BoolOp(self, node):
        # FIXME: use EVAL_BOOLOP
        if not isinstance(node.op, (ast.And, ast.Or)):
            return
        if len(node.values) < 2:
            return
        values = self.do_visit_list(node.values)
        constants = []
        for value in values:
            constant = self.get_constant(value)
            if constant is UNSET:
                break
            constants.append(constant)
        else:
            # all arguments are constant
            result = constants[0]
            if isinstance(node.op, ast.And):
                for constant in constants[1:]:
                    result = result and constant
            else:
                for constant in constants[1:]:
                    result = result or constant
            return self.new_constant(node, result)

        result = values
        for index, constant in enumerate(constants):
            test = constant
            if isinstance(node.op, ast.Or):
                test = not constant
            if test:
                result = values[index+1:]
            else:
                return self.new_constant(node, constant)
        if len(result) > 1:
            new_node = ast.BoolOp(node.op, result)
            return self.fix_node(node, new_node)
        else:
            return result[0]

    def visit_UnaryOp(self, node):
        eval_unaryop = EVAL_UNARYOP.get(node.op.__class__)
        if eval_unaryop is None:
            return
        operand = self.visit(node.operand)
        constant = self.get_constant(operand)
        if constant is UNSET:
            return ast.UnaryOp(node.op, operand)
        result = eval_unaryop(constant)
        if not self.config.valid_result(result):
            return
        return self.new_constant(node, result)

    def subscript_slice(self, node, value, lower, upper, step):
        lower_cst = self.get_constant(lower)
        upper_cst = self.get_constant(upper)
        step_cst = self.get_constant(step)
        if lower_cst is not UNSET and upper_cst is not UNSET and step_cst is not UNSET:
            if step_cst is not None and step_cst < 0:
                if lower_cst is not None:
                    last = lower_cst+1
                else:
                    last = len(value)
            else:
                if upper_cst is not None:
                    last = upper_cst
                else:
                    last = len(value)
            myslice = slice(lower_cst, upper_cst, step_cst)
            if not check_unicode_subscript(self.config, value[:last]):
                return UNSET
            return self.new_constant(node, value[myslice])

    def subscript_index(self, node, value, index):
        index_constant = self.get_constant(index)
        if index_constant is not UNSET:
            if not check_unicode_subscript(self.config, value[:index_constant]):
                return UNSET
            result = value[index_constant]
            return self.new_constant(node, result)

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        value_constant = self.get_constant(value)
        node_slice = node.slice
        if value_constant is not UNSET:
            if isinstance(node.slice, ast.Slice):
                lower = node.slice.lower
                if lower is not None:
                    lower = self.visit(lower)
                upper = node.slice.upper
                if upper is not None:
                    upper = self.visit(upper)
                step = node.slice.step
                if step is not None:
                    step = self.visit(step)
                new_node = self.subscript_slice(node, value_constant, lower, upper, step)
                if new_node is not UNSET:
                    return new_node
                node_slice = ast.Slice(lower, upper, step)
            if isinstance(node.slice, ast.Index):
                index = self.visit(node.slice.value)
                new_node = self.subscript_index(node, value_constant, index)
                if new_node is not UNSET:
                    return new_node
                node_slice = ast.Index(index)

        new_node = ast.Subscript(value=value, slice=node_slice, ctx=node.ctx)
        return self.fix_node(node, new_node)

    def visit_Compare(self, node):
        left = self.visit(node.left)
        left_constant = self.get_constant(left)
        comparators = self.do_visit_list(node.comparators)
        if left_constant is not UNSET and len(node.ops) == 1:
            op = node.ops[0]
            eval_compare = EVAL_COMPARE.get(op.__class__)
            if eval_compare is not None:
                assert len(comparators) == 1
                right = comparators[0]
                right_constant = self.get_constant(right)
                if right_constant is not UNSET:
                    if isinstance(op, ast.In):
                        result = eval_compare(right_constant, left_constant)
                    else:
                        result = eval_compare(left_constant, right_constant)
                    return self.new_constant(node, result)

        new_node = ast.Compare(
            left=left,
            ops=node.ops,
            comparators=comparators)
        return self.fix_node(node, new_node)

    def visit_If(self, node):
        new_test = self.visit(node.test)
        constant = self.get_constant(new_test)
        if constant is not UNSET:
            if constant:
                new_node = self.do_visit_list(node.body)
            else:
                new_node = self.do_visit_list(node.orelse)
            if not new_node:
                new_node = self.fix_node(node, ast.Pass())
            return new_node
        new_node = ast.If(
            test=new_test,
            body=node.body,
            orelse=node.orelse)
        return self.fix_node(node, new_node)

    def call_func(self, node, name, args):
        if name not in self.config.pure_functions:
            return UNSET
        check, eval_func = self.config.pure_functions[name]
        constants = self.get_constant_list(args)
        if constants is None:
            return UNSET
        if not check(self.config, constants):
            return UNSET
        result = eval_func(*constants)
        if not self.config.valid_result(result):
            return UNSET
        return self.new_constant(node, result)

    def call_method(self, node, func, args):
        obj_cst = self.get_constant(func.value)
        if obj_cst is UNSET:
            return UNSET
        attr = func.attr
        assert isinstance(attr, NATIVE_STR)

        obj_type = type(obj_cst)
        if obj_type not in self.config.pure_methods:
            return UNSET
        methods = self.config.pure_methods[obj_type]
        if attr not in methods:
            return UNSET
        check = methods[attr]

        constants = self.get_constant_list(args)
        if constants is None:
            return UNSET
        if not check(self.config, obj_cst, constants):
            return UNSET
        method = getattr(obj_cst, attr)
        result = method(*constants)
        return self.new_constant(node, result)

    def visit_Call(self, node):
        func = self.visit(node.func)
        args = self.do_visit_list(node.args)
        keywords = node.keywords
        starargs = node.starargs
        kwargs = node.kwargs
        # Don't support keywords, *args, **kw yet
        if (not keywords
        and not starargs
        and not kwargs):
            if isinstance(func, ast.Name):
                name = func.id
                new_node = self.call_func(node, name, args)
                if new_node is not UNSET:
                    return new_node
            elif isinstance(func, ast.Attribute) and isinstance(func.value, ast.Name):
                name = "%s.%s" % (func.value.id, func.attr)
                new_node = self.call_func(node, name, args)
                if new_node is not UNSET:
                    return new_node
            elif isinstance(func, ast.Attribute):
                new_node = self.call_method(node, func, args)
                if new_node is not UNSET:
                    return new_node
        new_node = ast.Call(func, args, keywords, starargs, kwargs)
        return self.fix_node(node, new_node)

