import unittest
from astoptimizer.compatibility import PYTHON3, u, b, ast
from astoptimizer import Config, parse_ast, Optimizer
import sys

class BaseTestCase(unittest.TestCase):
    TEXT_PASS = 'Module(body=[Pass()])'

    def text_num(self, value):
        return 'Module(body=[Expr(value=Num(n=%r))])' % value

    def text_name(self, value):
        return 'Module(body=[Expr(value=Name(id=%r, ctx=Load()))])' % value

    def text_bool(self, value):
        name = "True" if value else "False"
        return self.text_name(name)

    if PYTHON3:
        def text_bytes(self, value):
            return "Module(body=[Expr(value=Bytes(s=%r))])" % value
        def text_unicode(self, value):
            return "Module(body=[Expr(value=Str(s=%r))])" % value
    else:
        def text_bytes(self, value):
            return "Module(body=[Expr(value=Str(s=%r))])" % value
        def text_unicode(self, value):
            return "Module(body=[Expr(value=Str(s=%r))])" % value

    def text_native_str(self, value):
        return "Module(body=[Expr(value=Str(s=%r))])" % value

    def text_ast(self, code):
        tree = parse_ast(code)
        return ast.dump(tree)

    def check(self, code, expected, config=None):
        tree = parse_ast(code)
        tree = Optimizer(config).visit(tree)
        text = ast.dump(tree)
        self.assertEqual(text, expected)
        # Ensure that the tree is compilable to bytecode
        compile(tree, "test", "exec")

    def check_not_optimized(self, code, config=None):
        old_tree = parse_ast(code)
        old = ast.dump(old_tree)
        old = old.replace("UnaryOp(op=USub(), operand=Num(n=1))", "Num(n=-1)")
        new_tree = Optimizer(config).visit(old_tree)
        new = ast.dump(new_tree)
        self.assertEqual(new, old)


class TestOptimizer(BaseTestCase):
    def test_BinOp_int(self):
        self.check('1 + 1', self.text_num(2))
        self.check('10 - 1', self.text_num(9))
        self.check('2 * 3', self.text_num(6))
        self.check('21 // 5', self.text_num(4))
        self.check('10 ** 3', self.text_num(1000))
        self.check('1 << 5', self.text_num(32))
        self.check('32 >> 3', self.text_num(4))
        self.check('5 & 3', self.text_num(1))
        self.check('5 | 3', self.text_num(7))
        self.check('5 ^ 3', self.text_num(6))
        if PYTHON3:
            self.check('21 / 5', self.text_num(4.2))
        else:
            self.check_not_optimized('21 / 5')

        # complex operations
        self.check('1 + 2 * 3', self.text_num(7))

        # simple optimizations
        self.check('0 + x', self.text_name('x'))
        self.check_not_optimized('0.0 + x')
        self.check('x + 0', self.text_name('x'))
        self.check_not_optimized('x + 0.0')
        self.check('1 * x', self.text_name('x'))
        self.check_not_optimized('1.0 * x')
        self.check('x * 1', self.text_name('x'))
        self.check_not_optimized('x * 1.0')

        # test min_number / max_number limits
        config = Config()
        config.min_number = -10
        config.max_number = 10
        self.check('1 + 2', self.text_num(3), config)
        self.check_not_optimized('4 * 5', config)
        self.check_not_optimized('2 - 40', config)

    def test_BinOp_float(self):
        self.check('1.0 + 2', self.text_num(3.0))
        self.check('10 - 1.0', self.text_num(9.0))
        self.check('2.0 * 3.0', self.text_num(6.0))
        self.check('21.0 // 5.0', self.text_num(4.0))
        self.check('10.0 ** 3', self.text_num(1000.0))
        self.check_not_optimized('1.0 << 5')
        self.check_not_optimized('32.0 >> 3')
        self.check_not_optimized('5.0 & 3')
        self.check_not_optimized('5.0 | 3')
        self.check_not_optimized('5.0 ^ 3')

        # complex operations
        self.check('1.0 + 2 * 3.0', self.text_num(7.0))

        # test min_number / max_number limits
        config = Config()
        config.min_number = -10
        config.max_number = 10
        self.check('1.0 + 9', self.text_num(10.0), config)
        self.check_not_optimized('4.0 * 5', config)
        self.check_not_optimized('2.0 - 40', config)

    def test_BinOp_complex(self):
        self.check('1j + 2', self.text_num(1j + 2.0))
        self.check('10 - 1j', self.text_num(10.0 - 1j))
        self.check('2j * 3.0', self.text_num(6j))
        self.check('21.0 // 5.0', self.text_num(4.0))
        self.check('10.0 ** 3', self.text_num(1000.0))
        self.check_not_optimized('1.0 << 5')
        self.check_not_optimized('32.0 >> 3')
        self.check_not_optimized('5.0 & 3')
        self.check_not_optimized('5.0 | 3')
        self.check_not_optimized('5.0 ^ 3')

        # complex operations
        self.check('1.0 + 2 * 3.0', self.text_num(7.0))

        # test min_number / max_number limits
        config = Config()
        config.min_number = -10
        config.max_number = 10
        self.check('1j + 9j', self.text_num(10j), config)
        self.check_not_optimized('4j * 5', config)
        self.check_not_optimized('2j - 40j', config)

    def test_BinOp_string(self):
        self.check('%r * 3' % b("a"), self.text_bytes(b('aaa')))
        self.check('%r * 3' % (u('a'),), self.text_unicode(u('aaa')))
        self.check_not_optimized('"a" - "b"')
        self.check_not_optimized('"a" * "b"')
        self.check_not_optimized('"a" * 2.0')

        config = Config()
        config.max_string_length = 5
        self.check('"a"*5', self.text_unicode('aaaaa'), config)
        self.check_not_optimized('"a"*6', config)

    def test_BoolOp(self):
        # a and b constant
        self.check('True and True', self.text_bool(True))
        self.check('False or True', self.text_bool(True))
        self.check('13 or 14', self.text_num(13))

        # a constant
        self.check('True and x', self.text_name('x'))
        self.check('False and x', self.text_bool(False))
        self.check('True or x', self.text_bool(True))
        self.check('False or x', self.text_name('x'))

        # b constant
        self.check_not_optimized('x and True')
        self.check_not_optimized('x and False')
        self.check_not_optimized('x or True')
        self.check_not_optimized('x or False')

        # others
        self.check('4 and 5 and x and 6', self.text_ast('x and 6'))
        self.check('4 and 5 and ~6', self.text_num(-7))

    def test_UnaryOp(self):
        self.check('not True', self.text_bool(False))
        self.check('not 3', self.text_bool(False))
        self.check('not False', self.text_bool(True))
        self.check('not 0', self.text_bool(True))
        self.check_not_optimized('not x')

        self.check('~3', self.text_num(-4))
        self.check_not_optimized('~x')

        self.check('-(0.0)', self.text_num(-0.0))
        self.check('-(2)', self.text_num(-2))
        self.check('-(-(3))', self.text_num(3))

        config = Config()
        config.min_number = -10
        config.max_number = 10
        self.check('~9', self.text_num(-10), config)
        self.check_not_optimized('~10', config)

    def test_Subscript(self):
        self.check('%r[:3]' % b("abcdef"), self.text_bytes(b("abc")))
        if PYTHON3:
            self.check('%r[1]' % b("abcdef"), self.text_num(ord(b("b"))))
        else:
            self.check('%r[1]' % b("abcdef"), self.text_unicode(b("b")))
        self.check('"abcdef"[:3]', self.text_unicode("abc"))
        self.check('"abcdef"[1]', self.text_unicode("b"))
        self.check('%r[2]' % u("abc\U0010ffff-"), self.text_unicode(u("c")))
        self.check('%r[:3]' % u("abc\U0010ffff-"), self.text_unicode(u("abc")))
        self.check('%r[2::-1]' % u("abc\U0010ffff-"), self.text_unicode(u("cba")))
        self.check('%r[2:0:-1]' % u("abc\U0010ffff-"), self.text_unicode(u("cb")))
        if sys.version_info < (3,3):
            self.check_not_optimized('%r[4]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[-1]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[::-1]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[3::-1]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[3:0:-1]' % u("abc\U0010ffff-"))
        else:
            self.check('%r[3]' % u("abc\U0010ffff-"), self.text_unicode(u("\U0010ffff")))
            self.check('%r[-1]' % u("abc\U0010ffff-"), self.text_unicode("-"))

    def test_If(self):
        # FIXME: issue with "def func(): if 0: yield; return"
        # http://tomlee.co/2008/04/the-internals-of-python-generator-functions-in-the-ast/
        self.check('if 0: print("log")', self.TEXT_PASS)
        self.check('if 1: 3', self.text_num(3))
        self.check('if 2**2**2 - 16: 3', self.TEXT_PASS)

    def test_Compare(self):
        self.check('"abc" == "abc"', self.text_bool(True))
        self.check('1 != 2', self.text_bool(True))
        self.check('2 < 2', self.text_bool(False))
        self.check('2 <= 2', self.text_bool(True))
        self.check('2 > 2', self.text_bool(False))
        self.check('2 >= 2', self.text_bool(True))
        self.check('"xxx" in "abcdef"', self.text_bool(False))
        self.check('"bcd" in "abcdef"', self.text_bool(True))

    def test_Name(self):
        config = Config()
        config.constants['debug'] = 0
        self.check('debug', self.text_num(0), config)


class TestBuiltins(BaseTestCase):
    def test_Call_builtins(self):
        self.check('len("abc")', self.text_num(3))
        self.check('len("abcdef"[:2])', self.text_num(2))
        self.check('abs(-5)', self.text_num(5))
        self.check_not_optimized('abs("abc")')
        if sys.version_info >= (2,6):
            self.check('bin(5)', self.text_unicode('0b101'))
            self.check_not_optimized('bin(3.5)')
        self.check('bool(0)', self.text_bool(False))
        self.check('bool(3)', self.text_bool(True))
        self.check('bool("")', self.text_bool(False))
        self.check('bool("xxx")', self.text_bool(True))
        self.check('chr(65)', self.text_unicode('A'))
        self.check('float(6)', self.text_num(6.0))
        self.check('float(6.9)', self.text_num(6.9))
        self.check_not_optimized('float(1j)')
        self.check('int(6)', self.text_num(6))
        self.check('int(6.9)', self.text_num(6))
        self.check_not_optimized('int(1j)')
        self.check_not_optimized('oct(3.5)')
        self.check('ord(%r)' % b("A"), self.text_num(65))
        self.check('ord("A")', self.text_num(65))
        self.check_not_optimized('ord("abc")')
        self.check('pow(2, 3)', self.text_num(8))
        self.check('pow(2, 123, 4567)', self.text_num(197))
        self.check('pow(10.0, 3)', self.text_num(1000.0))
        self.check('repr(1)', self.text_unicode('1'))
        self.check("repr('abc')", self.text_unicode("'abc'"))
        self.check('str(1)', self.text_unicode('1'))
        self.check("str('abc')", self.text_unicode('abc'))

        if PYTHON3:
            self.check('oct(5)', self.text_unicode('0o5'))
            self.check("round(5)", self.text_num(5))
            self.check("round(5.5)", self.text_num(6))
            self.check("round(5, 1)", self.text_num(5))
            self.check("round(5.567, 1)", self.text_num(5.6))
        else:
            self.check('oct(5)', self.text_unicode('05'))
            self.check('long(6)', self.text_num(6))
            self.check('long(6.9)', self.text_num(6))
            self.check_not_optimized('long(1j)')
            self.check('unichr(65)', self.text_unicode(u('A')))
            self.check('repr(1L)', self.text_unicode('1L'))
            self.check('str(1L)', self.text_unicode('1'))
            self.check('unicode(1)', self.text_unicode(u('1')))
            self.check('unicode(1L)', self.text_unicode(u('1')))
            self.check("unicode('abc')", self.text_unicode(u('abc')))
            self.check_not_optimized("unicode('abc\xff')")
            self.check("round(5)", self.text_num(5.0))
            self.check("round(5.5)", self.text_num(6.0))
            self.check("round(5, 1)", self.text_num(5.0))
            self.check("round(5.567, 1)", self.text_num(5.6))

        config = Config()
        config.min_number = -10
        config.max_number = 10
        config.max_string_length = 5
        self.check('pow(2, 3)', self.text_num(8), config)
        self.check_not_optimized('pow(2, 10)', config)
        self.check('pow(2, 10, 7)', self.text_num(2), config)
        self.check_not_optimized('pow(2, 10, 42)', config)
        self.check('repr(12345)', self.text_unicode('12345'), config)
        self.check_not_optimized('repr(1234567)', config)
        self.check('str(12345)', self.text_unicode('12345'), config)
        self.check_not_optimized('str(1234567)', config)

    def test_Call_method(self):
        self.check('%r.decode("ascii")' % b("abc"), self.text_unicode(u('abc')))
        self.check('%r.encode("ascii")' % u('abc'), self.text_bytes(b('abc')))
        if sys.version_info >= (2, 7):
            self.check('(12345).bit_length()', self.text_num(14))

        if sys.version_info < (3,3):
            self.check_not_optimized('%r.encode("utf-8")' % u('\U0010ffff'))
        else:
            self.check(
                '%r.encode("utf-8")' % u('\U0010ffff'),
                self.text_bytes(b('\xf4\x8f\xbf\xbf')))



class TestFeatures(BaseTestCase):
    def test_platform(self):
        import platform

        config = Config('platform')
        self.check('sys.platform', self.text_native_str(sys.platform), config)
        self.check('platform.system()', self.text_native_str(platform.system()), config)

    def test_pythonenv(self):
        self.check_not_optimized('__debug__')
        self.check_not_optimized('sys.hexversion')
        self.check_not_optimized('sys.platform')
        if not PYTHON3:
            self.check_not_optimized('21 / 5')
        if sys.version_info < (3,3):
            self.check_not_optimized('%r[-1]' % u("abc\U0010ffff-"))

        config = Config('pythonenv')
        self.check('__debug__', self.text_bool(__debug__), config)
        self.check('sys.hexversion', self.text_num(sys.hexversion), config)
        self.check('sys.version[:3]', self.text_native_str(sys.version[:3]), config)
        if not PYTHON3:
            self.check('21 / 5', self.text_num(21 / 5), config)
        if sys.version_info < (3,3):
            if sys.maxunicode == 0x10ffff:
                self.check(
                    '%r[3]' % u("abc\U0010ffff-"),
                    self.text_unicode(u("\U0010ffff")),
                    config)
            else:
                self.check(
                    '%r[3]' % u("abc\U0010ffff-"),
                    self.text_unicode(u("\udbff")),
                    config)
                self.check(
                    '%r[4]' % u("abc\U0010ffff-"),
                    self.text_unicode(u("\udfff")),
                    config)
            self.check(
                '%r[-1]' % u("abc\U0010ffff-"),
                self.text_unicode(u("-")),
                config)

    def test_math(self):
        import math

        self.check('math.e', self.text_num(math.e))
        self.check('math.pi', self.text_num(math.pi))
        self.check('math.pow(2, 4)', self.text_num(16.0))
        self.check_not_optimized('math.pow(2j, 4)')
        self.check('math.exp(1.0)', self.text_num(math.e))
        self.check('math.log(1.0)', self.text_num(0.0))
        self.check('math.log(math.e)', self.text_num(1.0))
        self.check('math.sqrt(4.0)', self.text_num(2.0))

        # test min_number / max_number limits
        config = Config()
        config.min_number = -10
        config.max_number = 10
        self.check('math.pow(2, 3)', self.text_num(8.0), config)
        self.check_not_optimized('math.pow(2, 4)', config)
        self.check('math.exp(1.0)', self.text_num(math.e), config)
        self.check_not_optimized('math.exp(100.0)', config)


def main():
    unittest.main(module=__name__)

if __name__ == "__main__":
    main()

